(***********************************************
 * version with p g
 * semantics and lemma rules 
 **********************************************)
Require Export tm.
Require Export natutil.
Require Export map.
Require Export pg.
Require Import Arith.
Require Import EqNat.

(** sequential instruction cover *)
Definition icover (c : code) (i : index) (pg : spec) (sc : sins) : Prop :=
  match pg with (p,g) =>
  forall (s : state) (j : index) (pc : pcounter),
    CodeMod.maps_to c pc (sinsI sc)
    -> p c s j
    -> j < i
    -> (safe (sinsP sc) (c, s, j, pc) /\
       forall (c' : code) (s' : state) (j' : index) (pc' : pcounter), finish (sinsP sc) (c, s, j, pc) (c', s', j', pc') -> g c' s s' j')
  end.


(** function cover *)
Definition fcover (c : code) (i : index) (pg : spec) (is : iseq) : Prop :=
  match pg with (p,g) =>
  forall (s : state) (j : index) (pc : pcounter),
    lookupC c pc is
    -> p c s j
    -> j < i
    -> (safe (iseqP is) (c, s, j, pc) /\
       forall (c' : code) (s' : state) (j' : index) (pc' : pcounter), finish (iseqP is) (c, s, j, pc) (c', s', j', pc') -> g c' s s' j')
  end.

Lemma fcover_mono :
  forall (c : code) (i j : index) (pg : spec) (is : iseq),
    i> j -> fcover c i pg is -> fcover c j pg is.
Proof.
unfold fcover.
intros.
destruct pg as [p g].
intros.
apply H0; try assumption.
apply lt_trans with j; assumption.
Qed.

(** code heap specifition *)
Inductive psi : Type := 
  | empPS : psi
  | conjPS : lab -> spec -> psi -> psi.

Inductive lookupPS : psi -> lab -> spec -> Prop :=
  | lookupPS_eq :
    forall (ps : psi) (l l': lab) (pg : spec),
      l = l' -> lookupPS (conjPS l pg ps) l' pg
  | lookupC_neq :
    forall (ps : psi) (l l': lab) (pg pg': spec),
      l <> l' -> lookupPS ps l' pg' -> lookupPS (conjPS l pg ps) l' pg'.

Definition ccover  (c c': code) (i : index) (ps : psi) : Prop :=
  forall (l : lab) (pg : spec),
    lookupPS ps l pg
    -> exists is : iseq, (lookupC c' l is /\ fcover c i pg is).

(** well-specified sequential instructions *)
Definition wfsc (ps : psi) (pg : spec) (sc : sins) : Prop :=
  forall (c : code) (i : index),
    ccover c c i ps
    -> icover c (S i) pg sc.

(** well-specified functions *)
Definition wffunc (ps : psi) (pg : spec) (is : iseq) : Prop :=
  forall (c : code) (i : index),
    ccover c c i ps
    -> fcover c (S i) pg is. 

(** well-specified code heap *)
Definition wfch (ps ps' : psi) (c : code) : Prop :=
  (forall (l : lab) (pg : spec),
    lookupPS ps' l pg -> exists is : iseq, lookupC c l is) /\
  (forall (c' : code) (i : index),
    ccover c' c' i ps
    -> ccover c' c (S i) ps').

(** well-specified program *)
Definition wfprog (pg : spec) (c : code) (pc : pcounter) : Prop :=
  forall (i : index) (is : iseq),
    lookupC c pc is
    -> fcover c  i pg is.

Lemma call_rule :
  forall (ps : psi) (p : pred) (g : action) (l : lab),
    lookupPS ps l (psep p (rptox ra), gpsep g (rptox ra))
    -> monop p
    -> wfsc ps (psep p (rptox ra), gsep g (rptoxx ra)) (call l).
Proof.
unfold wfsc, ccover, icover, fcover.
intros.
induction j.
split.
apply safe_zero.
intros.
inversion H5.
destruct (H1 l (psep p (rptox ra), gpsep g (rptox ra)) H) as [is [I0 I1]].
clear H1 IHj.
assert (I2 : psep p (rptox ra) c s j).
assert (I3 : monop (psep p (rptox ra))).
apply monop_psep.
assumption.
apply monop_rptox.
unfold monop in I3.
apply I3 with (S j).
assumption.
apply gt_Sn_n.
assert (I3 : psep p (rptox ra) c (StateMod.put s (regR ra) (S pc)) j).
destruct I2 as [s1 [s2 [J0 [J1 J2]]]].
exists s1; exists (StateMod.put s2 (regR ra) (S pc)).
destruct J2 as [v [J2 J3]].
split.
apply StateMod.update_presv_disjUnion_right.
assumption.
exists v; assumption.
split.
assumption.
exists (S pc).
split.
apply StateMod.ext_mapsto.
intros.
apply StateMod.putA_presv_nidB.
apply J3; assumption.
assumption.
assert (I4 : j < i).
apply gt_S_n; assumption.
destruct (I1 (StateMod.put s (regR ra) (S pc)) j l I0 I3 I4) as [J0 J1].
split.
apply safe_call with is.
inversion H2.
assumption.
destruct H3 as [s1 [s2 [K [K1 [v [K2 K3]]]]]].
apply StateMod.disjUnion_presv_in_dom with s2 s1.
inversion K2.
exists v; assumption.
apply StateMod.disjUnion_commut; assumption.
assumption.
assumption.
intros.
assert (L : gpsep g (rptox ra) (w_cd w') (StateMod.put s (regR ra) (S pc)) (w_st w') (w_ind w')).
apply J1 with (w_pc w').
replace (w_cd w', w_st w', w_ind w', w_pc w') with w'.
assumption.
destruct w' as [[[cx sx] ix] pcx]; reflexivity.
assert (L0 : w_st w' (regR ra) = Some (w_pc w')).
apply ret_pc with is (c, StateMod.put s (regR ra) (S pc), j, l).
assumption.
assert (L1 : w_st w' (regR ra) = Some (S pc)).
destruct L as [s1 [s2 [s1' [L2 [L3 [L4 [v [L5 L6]]]]]]]].
assert (L8 : StateMod.in_dom (regR ra) s2).
exists v; inversion L5; reflexivity.
assert (L7 : StateMod.maps_to s2 (regR ra) (S pc)).
apply StateMod.disjUnion_in_right with s1 (StateMod.put s (regR ra) (S pc)).
assumption.
assumption.
apply StateMod.ext_mapsto.
assert (L9 : StateMod.maps_to (w_st w') (regR ra) (S pc)).
assert (L1 :  StateMod.disjUnion s2 s1' (w_st w')).
apply StateMod.disjUnion_commut; assumption.
destruct L1 as [M1 M2].
rewrite M2.
apply StateMod.in_left_in_merge.
assumption.
inversion L9; assumption.
rewrite L0 in L1; inversion L1.
reflexivity.
intros.
inversion H1; subst.
inversion H7. 
assert (K : is = is0).
apply lookupC_same with c l; assumption.
subst.
assert (K : gpsep g (rptox ra) c' (StateMod.put s (regR ra) (S pc)) s' j').
apply J1 with pc'; assumption.
destruct K as [s1 [s2 [s1' [K0 [K1 [K2 K3]]]]]].
destruct I2 as [s0 [s0' [L1 L2]]].
assert (K4 : StateMod.in_dom (regR ra) s).
destruct L2 as [L4 [v [L5 L6]]]. 
apply StateMod.disjUnion_presv_in_dom with s0' s0.
exists v; assumption.
apply StateMod.disjUnion_commut; assumption.
destruct K4 as [v K4].
exists s1; exists (StateMod.put s2 (regR ra) v); exists s1'; exists s2. 
split.
assert (K5 : s = StateMod.put (StateMod.put s (regR ra) (S pc)) (regR ra) v).
rewrite (StateMod.put_xx_update s (regR ra) (S pc) v).
apply StateMod.extensional_eqMap.
intros.
induction (StateMod.beq_A x (regR ra)).
subst x.
symmetry.
rewrite K4.
apply StateMod.put_get_eq.
symmetry.
apply StateMod.put_noninterfere.
intro.
apply b.
symmetry.
assumption.
rewrite K5.
apply StateMod.update_presv_disjUnion_right.
assumption.
destruct K3 as [v' [L L0]].
exists v'; assumption.
split.
assumption.
split.
assumption.
destruct K3 as [v' [L L0]].
exists v; exists v'.
unfold StateMod.sgltonM.
unfold StateMod.maps_to.
split.
split.
apply StateMod.put_get_eq.
intros.
apply StateMod.putA_presv_nidB.
apply L0; assumption.
assumption.
split.
assumption.
assumption.
Qed.

Definition cptr (l : lab) (pg : spec) : pred :=
  fun (c : code) (s : state) (i : index) =>
    exists is : iseq, lookupC c l is /\ fcover c i pg is.
 
Definition rcptr (r : reg) (pg : spec) : pred :=
  fun (c : code) (s : state) (i : index) =>
    exists l : lab, exists is : iseq, 
      lookupC c l is /\ s (regR r) = Some l /\ fcover c i pg is.

Lemma cr_rule :
  forall (ps : psi) (p : pred) (g : action) (r : reg),
    monop p
    -> wfsc ps (pand (psep p (rptox ra)) (rcptr r (psep p (rptox ra), gpsep g (rptox ra))), gsep g (rptoxx ra)) (cr r).
Proof.
unfold wfsc, icover.
intros.
induction j.
split.
apply safe_zero.
intros.
inversion H4.
clear IHj.
destruct H2 as [I [l [is [I0 [I1 I2]]]]].
assert (I3 : psep p (rptox ra) c s j).
assert (I4 : monop (psep p (rptox ra))).
apply monop_psep.
assumption.
apply monop_rptox.
unfold monop in I4.
apply I4 with (S j).
assumption.
apply gt_Sn_n.
unfold fcover in I2.
assert (I4 : psep p (rptox ra) c (StateMod.put s (regR ra) (S pc)) j).
destruct I3 as [s1 [s2 [J0 [J1 J2]]]].
exists s1; exists (StateMod.put s2 (regR ra) (S pc)).
destruct J2 as [v [J2 J3]].
split.
apply StateMod.update_presv_disjUnion_right.
assumption.
exists v; assumption.
split.
assumption.
exists (S pc).
split.
apply StateMod.ext_mapsto.
intros.
apply StateMod.putA_presv_nidB.
apply J3; assumption.
assumption.
assert (I5 : j < S j).
apply gt_Sn_n; assumption.
destruct (I2 (StateMod.put s (regR ra) (S pc)) j l I0 I4 I5) as [J0 J1].
split.
apply safe_cr with l is.
inversion H1.
assumption.
assumption.
assumption.
destruct I3 as [s1 [s2 [K [K1 [v [K2 K3]]]]]].
apply StateMod.disjUnion_presv_in_dom with s2 s1.
inversion K2.
exists v; assumption.
apply StateMod.disjUnion_commut; assumption.
assumption.
intros.
assert (L : gpsep g (rptox ra) (w_cd w') (StateMod.put s (regR ra) (S pc)) (w_st w') (w_ind w')).
apply J1 with (w_pc w').
replace (w_cd w', w_st w', w_ind w', w_pc w') with w'.
assumption.
destruct w' as [[[cx sx] ix] pcx]; reflexivity.
assert (L0 : w_st w' (regR ra) = Some (w_pc w')).
apply ret_pc with is (c, StateMod.put s (regR ra) (S pc), j, l).
assumption.
assert (L1 : w_st w' (regR ra) = Some (S pc)).
destruct L as [s1 [s2 [s1' [L2 [L3 [L4 [v [L5 L6]]]]]]]].
assert (L8 : StateMod.in_dom (regR ra) s2).
exists v; inversion L5; reflexivity.
assert (L7 : StateMod.maps_to s2 (regR ra) (S pc)).
apply StateMod.disjUnion_in_right with s1 (StateMod.put s (regR ra) (S pc)).
assumption.
assumption.
apply StateMod.ext_mapsto.
assert (L9 : StateMod.maps_to (w_st w') (regR ra) (S pc)).
assert (L1 :  StateMod.disjUnion s2 s1' (w_st w')).
apply StateMod.disjUnion_commut; assumption.
destruct L1 as [M1 M2].
rewrite M2.
apply StateMod.in_left_in_merge.
assumption.
inversion L9; assumption.
rewrite L0 in L1; inversion L1.
reflexivity.
intros.
inversion H2; subst.
inversion H6.
rewrite H10 in I1; inversion I1; subst l0.
assert (K : is = is0).
apply lookupC_same with c l; assumption.
subst is0.
assert (K : gpsep g (rptox ra) c' (StateMod.put s (regR ra) (S pc)) s' j').
apply J1 with pc'; assumption.
destruct K as [s1 [s2 [s1' [K0 [K1 [K2 K3]]]]]].
destruct I3 as [s0 [s0' [L1 L2]]].
assert (K4 : StateMod.in_dom (regR ra) s).
destruct L2 as [L4 [v [L5 L6]]]. 
apply StateMod.disjUnion_presv_in_dom with s0' s0.
exists v; assumption.
apply StateMod.disjUnion_commut; assumption.
destruct K4 as [v K4].
exists s1; exists (StateMod.put s2 (regR ra) v); exists s1'; exists s2. 
split.
assert (K5 : s = StateMod.put (StateMod.put s (regR ra) (S pc)) (regR ra) v).
rewrite (StateMod.put_xx_update s (regR ra) (S pc) v).
apply StateMod.extensional_eqMap.
intros.
induction (StateMod.beq_A x (regR ra)).
subst x.
symmetry.
rewrite K4.
apply StateMod.put_get_eq.
symmetry.
apply StateMod.put_noninterfere.
intro.
apply b.
symmetry.
assumption.
rewrite K5.
apply StateMod.update_presv_disjUnion_right.
assumption.
destruct K3 as [v' [L L0]].
exists v'; assumption.
split.
assumption.
split.
assumption.
destruct K3 as [v' [L L0]].
exists v; exists v'.
unfold StateMod.sgltonM.
unfold StateMod.maps_to.
split.
split.
apply StateMod.put_get_eq.
intros.
apply StateMod.putA_presv_nidB.
apply L0; assumption.
assumption.
split.
assumption.
assumption.
Qed.

Lemma sc_rule : 
  forall (ps : psi) (sc : sins),
    wfsc ps (fun (c : code) (s : state) (i : index) => condS sc s, fun (c : code) (s s' : state) (i : index) => nextS sc s s') sc.
Proof.
unfold wfsc, icover.
intros.
induction j.
split.
apply safe_zero.
intros.
inversion H3.
clear IHj.
split.
apply safe_sins.
inversion H0; assumption.
assumption.
intros.
inversion H3; subst sc.
assumption.
inversion H1.
inversion H1.
Qed.

Lemma ecpi_rule :
  forall (ps : psi) (l : lab) (pg : spec) (p : pred) (g : action) (sc : sins),
    lookupPS ps l pg
    -> wfsc ps ((pand (cptr l pg) p), g) sc
    -> wfsc ps (p, g) sc.
Proof.
unfold wfsc, icover.
intros.
apply H0 with i; try assumption.
split.
unfold cptr.
unfold ccover in H1.
destruct (H1 l pg H) as [is [I I0]].
exists is.
split.
assumption.
assert (J : j <i \/ j = i).
apply le_lt_or_eq.
apply lt_n_Sm_le; assumption.
destruct J as [J1 | J2].
apply fcover_mono with  i; assumption.
subst j; assumption.
assumption.
Qed.

Lemma weaki_rule :
  forall (ps : psi) (p p' : pred) (g g' : action) (sc : sins),
    wfsc ps (p, g) sc
    -> pimp p' p
    -> gimp g g'
    -> wfsc ps (p', g') sc.
Proof.
unfold wfsc, icover.
intros.
assert (I : p c s j).
apply H0; assumption.
destruct (H c i H2 s j pc H3 I H5) as [J J1].
split.
assumption.
intros.
apply H1.
apply J1 with pc'.
assumption.
Qed.

Lemma framei_rule :
  forall (ps : psi) (p p': pred) (g : action) (sc : sins),
    wfsc ps (p, g) sc
    -> monop p'
    -> wfsc ps (psep p p', gpsep g p') sc.
Proof.
unfold wfsc, icover.
intros.
destruct H3 as [s1 [s2 [I [I1 I2]]]].
destruct (H c i H1 s1 j pc H2 I1 H4) as [J J0].
split.
apply local_mono with s1 s2; assumption.
intros.
destruct (local_frame (sinsP sc) c c' s1 s2 s s' j j' pc pc' I J H3) as [s1' [K1 K2]]. 
exists s1; exists s2; exists s1'.
split.
assumption.
split.
assumption.
split.
apply J0 with pc'; assumption.
assert (L : w_cd (c, s1, j, pc) = w_cd (c', s1', j', pc')).
apply fin_c with (sinsP sc); assumption.
simpl in L.
subst c'.
unfold monop in H0.
apply H0 with j.
assumption.
assert (L0 : w_ind (c, s1, j, pc) > w_ind (c, s1', j', pc')).
apply fin_index_less with (sinsP sc); assumption.
simpl in L0; assumption.
Qed.

Lemma seq_rule :
  forall (ps : psi)  (p p': pred) (g g' : action) (sc : sins) (is : iseq),
    wfsc ps (p, g') sc
    -> wffunc ps (p', g) is
    -> pimp (pgtop p g') p'
    -> monop p
    -> monog g'
    -> wffunc ps (p, glink g' g) (sinsIS sc is).
Proof.
unfold wfsc, wffunc, icover, fcover.
intros.
induction j.
split.
apply safe_zero.
intros.
inversion H8.
inversion H16.
clear IHj.
inversion H5; subst.
destruct (H c i H4 s (S j) pc H12 H6 H7) as [I I0].
split.
apply safe_seq.
assumption.
assumption.
intros.
assert (J : w_cd (c, s, S j, pc) = w_cd w).
apply fin_c with (sinsP sc); assumption.
simpl in J.
assert (J0 : w_ind w < w_ind (c, s, S j, pc)).
apply fin_index_less with (sinsP sc); assumption.
assert (J1 : p' c (w_st w) (w_ind w)).
apply H1.
exists s;split.
apply H2 with (S j); assumption.
apply I0 with (w_pc w).
rewrite J at 2.
destruct w as[[[cx sx] ix] pcx].
assumption.
simpl in J0.
assert (J2 : w_ind w < S i).
apply lt_trans with (S j); assumption.
destruct (H0 c i H4 (w_st w) (w_ind w) (S pc) H13 J1 J2) as [I1 I2].  
assert (J3 : w_pc w = S (w_pc (c, s, S j, pc))).
apply fin_safe_sc_pc with sc; assumption.
simpl in J3.
rewrite <- J3 in I1.
rewrite J in I1.
destruct w as[[[cx sx] ix] pcx]; assumption.
intros.
inversion H8; subst.
destruct w as[[[c'' s''] j''] pc''].
assert (J : w_cd (c, s, S j, pc) = w_cd (c'', s'', j'', pc'')).
apply fin_c with (sinsP sc); assumption.
simpl in J; subst c''.
assert (J : w_cd (c, s'', j'', pc'') = w_cd (c', s', j', pc')).
apply fin_c with (iseqP is); assumption.
simpl in J; subst c'.
assert (J : w_pc (c, s'', j'', pc'') = S (w_pc (c, s, S j, pc))). 
apply fin_safe_sc_pc with sc; assumption.
simpl in J; subst pc''.
assert (J : w_ind (c, s'', j'', S pc) < w_ind (c, s, S j, pc)).
apply fin_index_less with (sinsP sc); assumption.
simpl in J.
assert (J0 : w_ind (c, s', j', pc') <  w_ind (c, s'', j'', S pc)).
apply fin_index_less with (iseqP is); assumption.
simpl in J0.
exists s''.
split.
unfold monog in H3.
apply H3 with j''.
apply I0 with (S pc); assumption.
exact J0.
assert (J1 : p' c s'' j'').
apply H1.
exists s.
split.
unfold monop in H2; apply H2 with (S j); assumption.
apply I0 with (S pc); assumption.
assert (J2 : j'' < S i).
apply lt_trans with (S j); assumption.
destruct (H0 c i H4 s'' j'' (S pc) H13 J1 J2) as [I1 I2].  
apply I2 with pc'; assumption.
Qed.

Definition monops (ps : psi) : Prop :=
  forall (l : lab) (p : pred) (g : action),
    lookupPS ps l (p, g)
    -> (monop p /\ monog g).

Lemma jmp_rule :
  forall (ps : psi) (pg : spec) (l : lab),
    lookupPS ps l pg
    -> monops ps
    -> wffunc ps pg (jinsIS (jmp l)).
Proof.
unfold wffunc, fcover.
intros.
destruct pg as [p g].
intros.
induction j.
split.
apply safe_zero.
intros.
inversion H5.
destruct (H1 l (p,g) H) as [is [I I0]].
clear IHj.
unfold fcover in I0.
assert (J : j < i).
apply lt_S_n; assumption.
assert (J0 : p c s j ).
destruct (H0 l p g H) as [I1 I2].
unfold monop in I1.
apply I1 with (S j).
assumption.
apply gt_Sn_n.
destruct (I0 s j l I J0 J) as [J1 J2]. 
split.
apply safe_jmp with is; assumption.
intros.
apply J2 with pc'.
inversion H5; subst.
assert (J3 : is = is0).
apply lookupC_same with c l; assumption.
subst is.
assumption.
Qed.

Lemma ret_rule : 
  forall (ps : psi),
    wffunc ps (rptox ra, gid) (jinsIS ret).
Proof.
unfold wffunc, fcover.
intros.
induction j.
split.
apply safe_zero.
intros.
inversion H3.
clear IHj.
split.
destruct H1 as [v [I I0]].
apply safe_ret.
assumption.
unfold StateMod.maps_to in I.
exists v.
assumption.
intros.
inversion H3.
unfold gid.
reflexivity.
Qed.

Lemma bgtz_rule :
  forall (ps : psi) (r : reg) (l : lab) (p p' : pred) (g g' : action) (is :iseq),
    lookupPS ps l (p', g')
    -> monop p
    -> monops ps
    -> wffunc ps (p, g) is
    -> wffunc ps (por (pand (rgtz r) p') (pand (rlez r) p), gor (pgtog (rgtz r) g') (pgtog (rlez r) g)) (binsIS (bgtz r l) is).
Proof.
unfold wffunc, fcover.
intros.
induction j.
split.
apply safe_zero.
intros.
inversion H7.
destruct H5 as [[I I0] | [I I0]].
destruct I as [v [J J0]].
destruct (H3 l (p',g') H) as [is' [I1 I2]].
clear IHj.
unfold fcover in I2.
assert (K : j < i).
apply lt_S_n; assumption.
assert (K0 : p' c s j ).
destruct (H1 l p' g' H) as [L1 L2].
unfold monop in L1.
apply L1 with (S j).
assumption.
apply gt_Sn_n.
destruct (I2 s j l I1 K0 K) as [L1 L2]. 
split.
unfold StateMod.maps_to in J.
apply safe_bgtz with is' v.
assumption.
assumption.
assumption.
intros.
assumption.
intros.
elim H5; assumption.
intros.
left.
inversion H5; subst.
assert (J3 : is' = is'0).
apply lookupC_same with c l; assumption.
subst is'0.
assert (J4 : w_cd (c, s, j, l) = w_cd (c', s', j', pc')).
apply fin_c with (iseqP is'); assumption.
simpl in J4; subst c'.
split.
exists v.
split; assumption.
apply L2 with pc'; assumption.
unfold StateMod.maps_to in J; rewrite H15 in J.
inversion J; subst v0.
elim H16; assumption.
destruct I as [v [J J0]].
assert (K : j < S i).
apply lt_trans with (S j).
apply gt_Sn_n.
assumption.
clear IHj.
assert (K0 : p c s j ).
unfold monop in H0.
apply H0 with (S j).
assumption.
apply gt_Sn_n.
inversion H4; subst.
destruct (H2 c i H3 s j (S pc) H11 K0 K) as [L1 L2].
destruct (H3 l (p',g') H) as [is' [I1 I2]].
split.
unfold StateMod.maps_to in J.
apply safe_bgtz with is' v.
assumption.
assumption.
assumption.
intros.
assert (M := le_not_gt v 0 J0).
elim M; assumption.
intros.
assumption.
intros.
right.
inversion H5; subst.
unfold StateMod.maps_to in J; rewrite J in H17.
inversion H17; subst v0.
assert (M := le_not_gt v 0 J0).
elim M; assumption.
assert (J4 : w_cd (c, s, j, (S pc)) = w_cd (c', s', j', pc')).
apply fin_c with (iseqP is); assumption.
simpl in J4; subst c'.
split.
exists v.
split; assumption.
apply L2 with pc'; assumption.
Qed.

Lemma ecpf_rule :
  forall (ps : psi) (l : lab) (pg : spec) (p : pred) (g : action) (is : iseq),
    lookupPS ps l pg
    -> wffunc ps ((pand (cptr l pg) p), g) is
    -> wffunc ps (p, g) is.
Proof.
unfold wffunc, fcover.
intros.
apply H0 with i; try assumption.
split.
unfold cptr.
unfold ccover in H1.
destruct (H1 l pg H) as [is0 [I I0]].
exists is0.
split.
assumption.
assert (J : j <i \/ j = i).
apply le_lt_or_eq.
apply lt_n_Sm_le; assumption.
destruct J as [J1 | J2].
apply fcover_mono with  i; assumption.
subst j; assumption.
assumption.
Qed.

Lemma weakf_rule :
  forall (ps : psi) (p p' : pred) (g g' : action) (is : iseq),
    wffunc ps (p, g) is
    -> pimp p' p
    -> gimp g g'
    -> wffunc ps (p', g') is.
Proof.
unfold wffunc, fcover.
intros.
assert (I : p c s j).
apply H0; assumption.
destruct (H c i H2 s j pc H3 I H5) as [J J1].
split.
assumption.
intros.
apply H1.
apply J1 with pc'.
assumption.
Qed.

Lemma framef_rule :
  forall (ps : psi) (p p': pred) (g : action) (is : iseq),
    wffunc ps (p, g) is
    -> monop p'
    -> wffunc ps (psep p p', gpsep g p') is.
Proof.
unfold wffunc, fcover.
intros.
destruct H3 as [s1 [s2 [I [I1 I2]]]].
destruct (H c i H1 s1 j pc H2 I1 H4) as [J J0].
split.
apply local_mono with s1 s2; assumption.
intros.
destruct (local_frame (iseqP is) c c' s1 s2 s s' j j' pc pc' I J H3) as [s1' [K1 K2]]. 
exists s1; exists s2; exists s1'.
split.
assumption.
split.
assumption.
split.
apply J0 with pc'; assumption.
assert (L : w_cd (c, s1, j, pc) = w_cd (c', s1', j', pc')).
apply fin_c with (iseqP is); assumption.
simpl in L.
subst c'.
unfold monop in H0.
apply H0 with j.
assumption.
assert (L0 : w_ind (c, s1, j, pc) > w_ind (c, s1', j', pc')).
apply fin_index_less with (iseqP is); assumption.
simpl in L0; assumption.
Qed.

Lemma heap_rule :
  forall (ps : psi) (c : code) (l : lab) (pg : spec) (is : iseq),
    lookupC c l is
    -> wffunc ps pg is
    -> wfch ps (conjPS l pg empPS) c.
Proof.
unfold wffunc, wfch.
intros.
split.
intros.
inversion H1; subst.
exists is; assumption.
inversion H8.
intros.
unfold ccover.
intros.
inversion H2; subst.
exists is.
split.
assumption.
apply H0; assumption.
inversion H9.
Qed.

Definition joinPS (ps ps' ps'' : psi) : Prop :=
  (forall (l : lab) (pg pg' : spec),
    lookupPS ps l pg -> lookupPS ps' l pg' -> pg = pg') /\
  (forall (l : lab) (pg : spec),
    lookupPS ps'' l pg <-> (lookupPS ps l pg \/ lookupPS ps' l pg)). 
Definition joinC (c c' c'' : code) : Prop :=
  (forall (l : lab) (is is' : iseq),
    lookupC c l is -> lookupC c' l is' -> is = is') /\
  (forall (l : lab) (is : iseq),
    lookupC c'' l is <-> (lookupC c l is \/ lookupC c' l is)).

Lemma joinPS_commut :
  forall (ps ps' ps'' : psi),
    joinPS ps ps' ps'' -> joinPS ps' ps ps''.
unfold joinPS.
intros.
destruct H as [I I0].
split.
intros.
assert (J : pg' = pg).
apply I with l; assumption.
subst pg';reflexivity.
intros.
destruct (I0 l pg) as [I1 I2].
split.
intros.
destruct (I1 H) as [K | K].
right; assumption.
left; assumption.
intros.
apply I2.
destruct H as [K | K].
right; assumption.
left; assumption.
Qed.

Lemma join_ccover :
  forall (c : code) (i : index) (ps ps' ps'': psi),
    ccover c c i ps''
    -> joinPS ps ps' ps''
    -> ccover c c i ps.
Proof.
unfold ccover.
intros.
destruct H0 as [I I0].
destruct  (I0 l pg) as [I1 I2].
apply H.
apply I2.
left; assumption.
Qed.

Lemma link_rule : 
  forall (ps ps1 ps' ps1' ps'' ps1'' : psi) (c c' c'': code),
    wfch ps ps1 c 
    -> wfch ps' ps1' c'
    -> joinPS ps ps' ps''
    -> joinPS ps1 ps1' ps1''
    -> joinC c c' c''
    -> wfch ps'' ps1'' c''.
Proof.
unfold wfch.
intros.
destruct H as [I I0].
destruct H0 as [I1 I2].
destruct H3 as [I3 I4].
destruct H1 as [I5 I6].
destruct H2 as [I7 I8]. 
split.
intros.
destruct (I8 l pg) as [K K0].
destruct (K H) as [L | L].
destruct (I l pg L) as [is M].
exists is.
destruct (I4 l is) as [N N0].
apply N0.
left; assumption.
destruct (I1 l pg L) as [is M].
exists is.
destruct (I4 l is) as [N N0].
apply N0.
right; assumption.
intros.
unfold ccover.
intros.
destruct (I8 l pg) as [J J0].
destruct (J H0) as [K | K].
assert (L : ccover c'0 c'0 i ps).
apply join_ccover with ps' ps''.
assumption.
split; assumption.
assert (M := I0 c'0 i L).
unfold ccover in M.
destruct (M l pg K) as [is [M0 M1]].
exists is.
split.
destruct (I4 l is) as [R R1].
apply R1.
left; assumption.
assumption.
assert (L : ccover c'0 c'0 i ps').
apply join_ccover with ps ps''.
assumption.
apply joinPS_commut.
split; assumption.
assert (M := I2 c'0 i L).
unfold ccover in M.
destruct (M l pg K) as [is [M0 M1]].
exists is.
split.
destruct (I4 l is) as [R R1].
apply R1.
right; assumption.
assumption.
Qed.

Lemma prog_rule :
  forall (ps : psi) (c : code) (pg : spec) (pc : pcounter) (is : iseq),
    lookupC c pc is
    -> wfch ps ps c
    -> wffunc ps pg is
    -> wfprog pg c pc.
Proof.
unfold wfch, wffunc, wfprog.
intros.
assert (I : is = is0).
apply lookupC_same with c pc; assumption.
subst is0.
assert (I : forall i : index, ccover c c i ps).
intro j.
destruct H0 as [I I0].
induction j.
unfold ccover.
intros.
destruct (I l pg0) as [is0 I1].
assumption.
exists is0.
split.
assumption.
unfold fcover.
destruct pg0 as [px gx].
intros.
inversion H5.
apply I0; assumption.
induction i.
unfold fcover; destruct pg as [px gx]; intros.
inversion H5.
apply H1.
apply I.
Qed.


